@@ -131,8 +131,9 @@ void DeclarationBuilder::getVariableIdentifier(VariableAst* node,
 }
 
 ReferencedTopDUContext DeclarationBuilder::build(const IndexedString& url, AstNode* node,
-        ReferencedTopDUContext updateContext)
+                                                 const ReferencedTopDUContext& updateContext_)
 {
+    ReferencedTopDUContext updateContext(updateContext_);
     //Run DeclarationBuilder twice, to find uses of declarations that are
     //declared after the use. ($a = new Foo; class Foo {})
     {
@@ -191,6 +192,12 @@ void DeclarationBuilder::visitClassDeclarationStatement(ClassDeclarationStatemen
     closeType();
     closeDeclaration();
     m_upcomingClassVariables.clear();
+
+    QString className = classDec->prettyName().str();
+
+    if (isReservedClassName(className)) {
+        reportError(i18n("Cannot use '%1' as class name as it is reserved", className), node->className);
+    }
 }
 
 void DeclarationBuilder::visitInterfaceDeclarationStatement(InterfaceDeclarationStatementAst *node)
@@ -200,6 +207,12 @@ void DeclarationBuilder::visitInterfaceDeclarationStatement(InterfaceDeclaration
     DeclarationBuilderBase::visitInterfaceDeclarationStatement(node);
     closeType();
     closeDeclaration();
+
+    QString interfaceName = interfaceDec->prettyName().str();
+
+    if (isReservedClassName(interfaceName)) {
+        reportError(i18n("Cannot use '%1' as class name as it is reserved", interfaceName), node->interfaceName);
+    }
 }
 
 void DeclarationBuilder::visitTraitDeclarationStatement(TraitDeclarationStatementAst * node)
@@ -210,6 +223,12 @@ void DeclarationBuilder::visitTraitDeclarationStatement(TraitDeclarationStatemen
     closeType();
     closeDeclaration();
     m_upcomingClassVariables.clear();
+
+    QString traitName = traitDec->prettyName().str();
+
+    if (isReservedClassName(traitName)) {
+        reportError(i18n("Cannot use '%1' as class name as it is reserved", traitName), node->traitName);
+    }
 }
 
 ClassDeclaration* DeclarationBuilder::openTypeDeclaration(IdentifierAst* name, ClassDeclarationData::ClassType type)
@@ -580,31 +599,6 @@ void DeclarationBuilder::declareClassMember(DUContext *parentCtx, AbstractType::
 
     DUChainWriteLocker lock(DUChain::lock());
 
-    // check for redeclaration of private or protected stuff
-    {
-        // only interesting context might be the class context when we are inside a method
-        DUContext *ctx = currentContext()->parentContext();
-        foreach ( Declaration* dec, parentCtx->findDeclarations(identifier) ) {
-            if ( ClassMemberDeclaration* cdec = dynamic_cast<ClassMemberDeclaration*>(dec) ) {
-                if ( cdec->accessPolicy() == Declaration::Private && cdec->context() != ctx ) {
-                    reportError(i18n("Cannot redeclare private property %1 from this context.",
-                                        cdec->toString()), node);
-                    return;
-                } else if ( cdec->accessPolicy() == Declaration::Protected
-                            && cdec->context() != ctx
-                            && ( !ctx || !ctx->imports(cdec->context()) ) ) {
-                    reportError(i18n("Cannot redeclare protected property %1 from this context.",
-                                        cdec->toString()), node);
-                    return;
-                }
-                if ( cdec->abstractType()->indexed() == type->indexed() ) {
-                    encounter(dec);
-                    return;
-                }
-            }
-        }
-    }
-
     // this member should be public and non-static
     m_currentModifers = ModifierPublic;
     injectContext(parentCtx);
@@ -619,6 +613,31 @@ void DeclarationBuilder::declareClassMember(DUContext *parentCtx, AbstractType::
 
 void DeclarationBuilder::visitConstantDeclaration(ConstantDeclarationAst *node)
 {
+    DUChainWriteLocker lock(DUChain::lock());
+    if (m_reportErrors) {
+        // check for redeclarations
+        foreach(Declaration * dec, currentContext()->findLocalDeclarations(identifierForNode(node->identifier).first(), startPos(node->identifier)))
+        {
+            if (wasEncountered(dec) && !dec->isFunctionDeclaration() && dec->abstractType()->modifiers() & AbstractType::ConstModifier) {
+                reportRedeclarationError(dec, node->identifier);
+                break;
+            }
+        }
+    }
+    ClassMemberDeclaration* dec = openDefinition<ClassMemberDeclaration>(identifierForNode(node->identifier), m_editor->findRange(node->identifier));
+    {
+        DUChainWriteLocker lock(DUChain::lock());
+        dec->setAccessPolicy(Declaration::Public);
+        dec->setStatic(true);
+        dec->setKind(Declaration::Instance);
+    }
+    DeclarationBuilderBase::visitConstantDeclaration(node);
+    closeDeclaration();
+}
+
+void DeclarationBuilder::visitClassConstantDeclaration(ClassConstantDeclarationAst *node)
+{
+    DUChainWriteLocker lock;
     if (m_reportErrors) {
         // Check for constants in traits
         if (isMatch(currentDeclaration(), ClassDeclarationType)) {
@@ -630,8 +649,13 @@ void DeclarationBuilder::visitConstantDeclaration(ConstantDeclarationAst *node)
             }
         }
 
+        // check for 'class' constant
+        if (identifierForNode(node->identifier).toString().toLower() == QLatin1String("class"))
+        {
+            reportError(i18n("A class constant must not be called 'class'; it is reserved for class name fetching"), node);
+        }
+
         // check for redeclarations
-        DUChainWriteLocker lock(DUChain::lock());
         foreach(Declaration * dec, currentContext()->findLocalDeclarations(identifierForNode(node->identifier).first(), startPos(node->identifier)))
         {
             if (wasEncountered(dec) && !dec->isFunctionDeclaration() && dec->abstractType()->modifiers() & AbstractType::ConstModifier) {
@@ -640,14 +664,13 @@ void DeclarationBuilder::visitConstantDeclaration(ConstantDeclarationAst *node)
             }
         }
     }
-    ClassMemberDeclaration* dec = openDefinition<ClassMemberDeclaration>(node->identifier, node->identifier);
-    {
-        DUChainWriteLocker lock(DUChain::lock());
-        dec->setAccessPolicy(Declaration::Public);
-        dec->setStatic(true);
-        dec->setKind(Declaration::Instance);
-    }
-    DeclarationBuilderBase::visitConstantDeclaration(node);
+    ClassMemberDeclaration* dec = openDefinition<ClassMemberDeclaration>(identifierForNode(node->identifier), m_editor->findRange(node->identifier));
+    dec->setAccessPolicy(Declaration::Public);
+    dec->setStatic(true);
+    dec->setKind(Declaration::Instance);
+    lock.unlock();
+
+    DeclarationBuilderBase::visitClassConstantDeclaration(node);
     closeDeclaration();
 }
 
@@ -662,7 +685,6 @@ void DeclarationBuilder::visitTraitAliasStatement(TraitAliasStatementAst *node)
     }
 
     lock.unlock();
-
     DeclarationBuilderBase::visitTraitAliasStatement(node);
 }
 
@@ -674,6 +696,8 @@ void DeclarationBuilder::createTraitAliasDeclarations(TraitAliasStatementAst *no
     QualifiedIdentifier alias;
     if (node->aliasIdentifier) {
         alias = identifierPairForNode(node->aliasIdentifier).second;
+    } else if (node->aliasNonModifierIdentifier) {
+        alias = identifierPairForNode(node->aliasNonModifierIdentifier).second;
     } else {
         alias = original;
     }
@@ -683,9 +707,14 @@ void DeclarationBuilder::createTraitAliasDeclarations(TraitAliasStatementAst *no
         TraitMethodAliasDeclaration* newdec;
 
         // no existing declaration found, create one
-        if (node->aliasIdentifier) {
-            newdec = openDefinition<TraitMethodAliasDeclaration>(alias, m_editor->findRange(node->aliasIdentifier));
-            newdec->setPrettyName(identifierPairForNode(node->aliasIdentifier).first);
+        if (node->aliasIdentifier || node->aliasNonModifierIdentifier) {
+            if (node->aliasIdentifier) {
+                newdec = openDefinition<TraitMethodAliasDeclaration>(alias, m_editor->findRange(node->aliasIdentifier));
+                newdec->setPrettyName(identifierPairForNode(node->aliasIdentifier).first);
+            } else {
+                newdec = openDefinition<TraitMethodAliasDeclaration>(alias, m_editor->findRange(node->aliasNonModifierIdentifier));
+                newdec->setPrettyName(identifierPairForNode(node->aliasNonModifierIdentifier).first);
+            }
             newdec->setAccessPolicy(olddec->accessPolicy());
             openAbstractType(olddec->abstractType());
             if (node->modifiers) {
@@ -697,13 +726,15 @@ void DeclarationBuilder::createTraitAliasDeclarations(TraitAliasStatementAst *no
                     newdec->setAccessPolicy(Declaration::Private);
                 }
 
+                if (node->modifiers->modifiers & ModifierAbstract) {
+                    reportError(i18n("Cannot use 'abstract' as method modifier"), node->modifiers, IProblem::Error);
+                }
                 if (node->modifiers->modifiers & ModifierFinal) {
                     reportError(i18n("Cannot use 'final' as method modifier"), node->modifiers, IProblem::Error);
                 }
                 if (node->modifiers->modifiers & ModifierStatic) {
                     reportError(i18n("Cannot use 'static' as method modifier"), node->modifiers, IProblem::Error);
                 }
-
             }
         } else {
             CursorInRevision cursor = m_editor->findRange(node->importIdentifier).start;
@@ -711,6 +742,26 @@ void DeclarationBuilder::createTraitAliasDeclarations(TraitAliasStatementAst *no
             newdec->setPrettyName(identifierPairForNode(node->importIdentifier->methodIdentifier).first);
             newdec->setAccessPolicy(olddec->accessPolicy());
             openAbstractType(olddec->abstractType());
+
+            if (node->modifiers) {
+                if (node->modifiers->modifiers & ModifierPublic) {
+                    newdec->setAccessPolicy(Declaration::Public);
+                } else if (node->modifiers->modifiers & ModifierProtected) {
+                    newdec->setAccessPolicy(Declaration::Protected);
+                } else if (node->modifiers->modifiers & ModifierPrivate) {
+                    newdec->setAccessPolicy(Declaration::Private);
+                }
+
+                if (node->modifiers->modifiers & ModifierAbstract) {
+                    reportError(i18n("Cannot use 'abstract' as method modifier"), node->modifiers, IProblem::Error);
+                }
+                if (node->modifiers->modifiers & ModifierFinal) {
+                    reportError(i18n("Cannot use 'final' as method modifier"), node->modifiers, IProblem::Error);
+                }
+                if (node->modifiers->modifiers & ModifierStatic) {
+                    reportError(i18n("Cannot use 'static' as method modifier"), node->modifiers, IProblem::Error);
+                }
+            }
         }
         newdec->setKind(Declaration::Type);
         newdec->setAliasedDeclaration(IndexedDeclaration(olddec));
@@ -743,7 +794,7 @@ void DeclarationBuilder::createTraitAliasDeclarations(TraitAliasStatementAst *no
 
 void DeclarationBuilder::visitParameterList(ParameterListAst* node)
 {
-    PushValue<ParameterAst*> push(m_functionDeclarationPreviousArgument, 0);
+    PushValue<ParameterAst*> push(m_functionDeclarationPreviousArgument, nullptr);
 
     DeclarationBuilderBase::visitParameterList(node);
 }
@@ -758,11 +809,21 @@ void DeclarationBuilder::visitParameter(ParameterAst *node)
         funDec->addDefaultParameter(IndexedString(symbol));
         if (node->isVariadic != -1) {
             reportError(i18n("Variadic parameter cannot have a default value"), node->defaultValue);
-        } else if ( node->parameterType && node->parameterType->objectType && symbol.compare(QLatin1String("null"), Qt::CaseInsensitive) != 0 ) {
+        } else if (node->parameterType && node->parameterType->typehint && isClassTypehint(node->parameterType->typehint, m_editor) &&
+                symbol.compare(QLatin1String("null"), Qt::CaseInsensitive) != 0) {
             reportError(i18n("Default value for parameters with a class type hint can only be NULL."), node->defaultValue);
+        } else if (node->parameterType && node->parameterType->typehint && node->parameterType->typehint->genericType &&
+                symbol.compare(QLatin1String("null"), Qt::CaseInsensitive) != 0) {
+            NamespacedIdentifierAst* typehintNode = node->parameterType->typehint->genericType;
+            const KDevPG::ListNode< IdentifierAst* >* it = typehintNode->namespaceNameSequence->back();
+            QString typehintName = m_editor->parseSession()->symbol(it->element);
+
+            if (typehintName.compare(QLatin1String("object"), Qt::CaseInsensitive) == 0) {
+                reportError(i18n("Default value for parameters with an object type can only be NULL."), node->defaultValue);
+            }
         }
-    } else if ( !node->defaultValue && funDec->defaultParametersSize() ) {
-        reportError(i18n("Following parameters must have a default value assigned."), node);
+    } else {
+        funDec->addDefaultParameter(IndexedString{});
     }
     {
         // create variable declaration for argument
@@ -775,6 +836,16 @@ void DeclarationBuilder::visitParameter(ParameterAst *node)
 
     DeclarationBuilderBase::visitParameter(node);
 
+    if (node->parameterType && node->parameterType->typehint && isClassTypehint(node->parameterType->typehint, m_editor)) {
+        NamespacedIdentifierAst* typehintNode = node->parameterType->typehint->genericType;
+        const KDevPG::ListNode< IdentifierAst* >* it = typehintNode->namespaceNameSequence->back();
+        QString className = m_editor->parseSession()->symbol(it->element);
+
+        if (isReservedClassName(className)) {
+            reportError(i18n("Cannot use '%1' as class name as it is reserved", className), typehintNode);
+        }
+    }
+
     if (m_functionDeclarationPreviousArgument && m_functionDeclarationPreviousArgument->isVariadic != -1) {
         reportError(i18n("Only the last parameter can be variadic."), m_functionDeclarationPreviousArgument);
     }
@@ -802,6 +873,19 @@ void DeclarationBuilder::visitFunctionDeclarationStatement(FunctionDeclarationSt
     closeType();
     closeDeclaration();
 }
+
+void DeclarationBuilder::visitReturnType(ReturnTypeAst* node) {
+    if (node->typehint && isClassTypehint(node->typehint, m_editor)) {
+        NamespacedIdentifierAst* typehintNode = node->typehint->genericType;
+        const KDevPG::ListNode< IdentifierAst* >* it = typehintNode->namespaceNameSequence->back();
+        QString className = m_editor->parseSession()->symbol(it->element);
+
+        if (isReservedClassName(className)) {
+            reportError(i18n("Cannot use '%1' as class name as it is reserved", className), typehintNode);
+        }
+    }
+}
+
 void DeclarationBuilder::visitClosure(ClosureAst* node)
 {
     setComment(formatComment(node, editor()));
@@ -1100,7 +1184,98 @@ void DeclarationBuilder::visitAssignmentExpressionEqual(AssignmentExpressionEqua
 {
     DeclarationBuilderBase::visitAssignmentExpressionEqual(node);
 
-    if ( !m_findVariable.identifier.isEmpty() && currentAbstractType()) {
+    bool alreadyDeclared = false;
+
+    // check for already declared class members
+    if (!m_findVariable.identifier.isEmpty() && !m_findVariable.parentIdentifier.isEmpty()) {
+        DUContext* ctx = getClassContext(m_findVariable.parentIdentifier, currentContext());
+        if (ctx) {
+            DUChainReadLocker lock(DUChain::lock());
+            ifDebug(qCDebug(DUCHAIN) << "checking if already declared: " << m_findVariable.identifier.toString();)
+            QList<Declaration*> decs;
+            foreach ( Declaration* dec, currentContext()->findDeclarations(m_findVariable.identifier) ) {
+                if ( dec->kind() == Declaration::Kind::Instance) {
+                    if (dec->range() == editorFindRange(m_findVariable.node)) {
+                        // Don't reuse previous declarations as the type might have changed.
+                        continue;
+                    }
+
+                    ClassMemberDeclaration *classDec = dynamic_cast<ClassMemberDeclaration*>(dec);
+
+                    if (classDec && classDec->accessPolicy() == Declaration::Private) {
+                        if (currentContext()->parentContext() == dec->context()) {
+                            decs << dec;
+                            ifDebug(qCDebug(DUCHAIN) << "found class property:" << dec->toString();)
+                        }
+                    } else {
+                        decs << dec;
+                        ifDebug(qCDebug(DUCHAIN) << "found class property:" << dec->toString();)
+                    }
+                }
+            }
+            lock.unlock();
+
+            if (!decs.isEmpty()) {
+                Declaration * dec = decs.last();
+
+                if (dec) {
+                    encounter(dec);
+                    alreadyDeclared = true;
+                    IntegralType::Ptr type = dec->type<IntegralType>();
+
+                    if (type && type->dataType() == IntegralType::TypeNull) {
+                        DUChainWriteLocker wlock;
+                        dec->setAbstractType(currentAbstractType());
+                    }
+                }
+            }
+
+            lock.lock();
+            if (decs.isEmpty() && (!currentContext()->parentContext() || !currentContext()->parentContext()->imports(ctx))) {
+                ifDebug(qCDebug(DUCHAIN) << "nothing found in current context, look in class context of parent";)
+                QList<Declaration*> decs;
+                foreach ( Declaration* dec, ctx->findDeclarations(m_findVariable.identifier) ) {
+                    if ( dec->kind() == Declaration::Kind::Instance) {
+                        if (dec->range() == editorFindRange(m_findVariable.node)) {
+                            // Don't reuse previous declarations as the type might have changed.
+                            continue;
+                        }
+
+                        ifDebug(qCDebug(DUCHAIN) << "found class property:" << dec->toString();)
+                        decs << dec;
+                    }
+                }
+                lock.unlock();
+
+                if (!decs.isEmpty()) {
+                    Declaration * dec = decs.last();
+
+                    if (dec) {
+                        alreadyDeclared = true;
+                        IntegralType::Ptr type = dec->type<IntegralType>();
+
+                        // check for redeclaration of private or protected stuff
+                        DUContext *parentCtx = currentContext()->parentContext();
+                        ClassMemberDeclaration *classDec = dynamic_cast<ClassMemberDeclaration*>(dec);
+
+                        if (classDec && classDec->accessPolicy() == Declaration::Private && classDec->context() != parentCtx) {
+                            reportError(i18n("Cannot access private property %1",
+                                                classDec->toString()), m_findVariable.node);
+                        } else if (classDec && classDec->accessPolicy() == Declaration::Protected && classDec->context() != parentCtx) {
+                            reportError(i18n("Cannot access protected property %1",
+                                                classDec->toString()), m_findVariable.node);
+                        } else if (type && type->dataType() == IntegralType::TypeNull) {
+                            DUChainWriteLocker wlock;
+                            dec->setAbstractType(currentAbstractType());
+                        }
+                    }
+                }
+            }
+        }
+    }
+
+    if ( !alreadyDeclared && !m_findVariable.identifier.isEmpty() && currentAbstractType()) {
+        ifDebug(qCDebug(DUCHAIN) << "not yet declared: " << m_findVariable.identifier.toString();)
         //create new declaration assignments to not-yet declared variables and class members
 
         AbstractType::Ptr type;
@@ -1133,9 +1308,18 @@ void DeclarationBuilder::visitFunctionCall(FunctionCallAst* node)
         DeclarationPointer dec;
         if ( node->stringFunctionName ) {
             dec = findDeclarationImport(FunctionDeclarationType, node->stringFunctionName);
+
+            if (!dec) {
+                dec = findDeclarationImport(FunctionDeclarationType, node->stringFunctionName, GlobalScope);
+            }
         } else if ( node->stringFunctionNameOrClass ) {
             id = identifierForNamespace(node->stringFunctionNameOrClass, m_editor);
             dec = findDeclarationImport(FunctionDeclarationType, id);
+
+            if (!dec) {
+                id.setExplicitlyGlobal(true);
+                dec = findDeclarationImport(FunctionDeclarationType, id);
+            }
         } else {
             ///TODO: node->varFunctionName
         }
@@ -1197,7 +1381,7 @@ void DeclarationBuilder::visitFunctionCall(FunctionCallAst* node)
 
 void DeclarationBuilder::visitFunctionCallParameterList(FunctionCallParameterListAst* node)
 {
-    PushValue<FunctionCallParameterListElementAst*> push(m_functionCallPreviousArgument, 0);
+    PushValue<FunctionCallParameterListElementAst*> push(m_functionCallPreviousArgument, nullptr);
     PushValue<int> pos(m_functionCallParameterPos, 0);
 
     DeclarationBuilderBase::visitFunctionCallParameterList(node);
@@ -1431,11 +1615,15 @@ void DeclarationBuilder::visitUseNamespace(UseNamespaceAst* node)
 
     ///TODO: case insensitive!
     QualifiedIdentifier qid = identifierForNamespace(node->identifier, m_editor);
-    ///TODO: find out why this must be done (see mail to kdevelop-devel on jan 18th 2011)
-    qid.setExplicitlyGlobal( false );
 
     DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, qid);
 
+    if (!dec && !qid.explicitlyGlobal()) {
+        QualifiedIdentifier globalQid = qid;
+        globalQid.setExplicitlyGlobal(true);
+        dec = findDeclarationImport(ClassDeclarationType, globalQid);
+    }
+
     if (dec)
     {
         // Check for a name conflict
@@ -1456,6 +1644,9 @@ void DeclarationBuilder::visitUseNamespace(UseNamespaceAst* node)
     }
     else
     {
+        // NamespaceAliasDeclarations can't use a global import identifier
+        qid.setExplicitlyGlobal(false);
+
         NamespaceAliasDeclaration* decl = openDefinition<NamespaceAliasDeclaration>(id.second,
                                                                                     m_editor->findRange(idNode));
         decl->setImportIdentifier( qid );
@@ -1463,6 +1654,23 @@ void DeclarationBuilder::visitUseNamespace(UseNamespaceAst* node)
         decl->setKind(Declaration::NamespaceAlias);
     }
     closeDeclaration();
+
+    if (node->aliasIdentifier) {
+        QString aliasName = m_editor->parseSession()->symbol(node->aliasIdentifier);
+
+        if (isReservedClassName(aliasName)) {
+            reportError(i18n("Cannot use %1 as %2 because '%2' is a special class name", qid.toString(), aliasName), node->aliasIdentifier);
+        }
+    }
+}
+
+void DeclarationBuilder::visitVarExpression(VarExpressionAst* node)
+{
+    DeclarationBuilderBase::visitVarExpression(node);
+
+    if (node->isGenerator != -1 && currentContext()->type() != DUContext::Other) {
+        reportError(i18n("The 'yield' expression can only be used inside a function"), node);
+    }
 }
 
 void DeclarationBuilder::updateCurrentType()
@@ -1503,4 +1711,18 @@ void DeclarationBuilder::encounter(Declaration* dec)
     }
 }
 
+bool DeclarationBuilder::isReservedClassName(QString className)
+{
+    return className.compare(QLatin1String("string"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("bool"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("int"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("float"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("iterable"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("object"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("null"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("true"), Qt::CaseInsensitive) == 0
+            || className.compare(QLatin1String("false"), Qt::CaseInsensitive) == 0;
+}
+
+
 }
