@@ -190,19 +190,35 @@ void ExpressionVisitor::visitVariable(VariableAst* node)
     }
 }
 
+void ExpressionVisitor::visitVarExpression(VarExpressionAst *node)
+{
+    DefaultVisitor::visitVarExpression(node);
+    if (node->isGenerator != -1) {
+        DeclarationPointer generatorDecl = findDeclarationImport(ClassDeclarationType, QualifiedIdentifier("generator"));
+
+        if (generatorDecl) {
+            m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed)));
+            if (hasCurrentClosureReturnType()) {
+                FunctionType::Ptr closureType = currentClosureReturnType().cast<FunctionType>();
+                closureType->setReturnType(generatorDecl->abstractType());
+            }
+        }
+    }
+}
+
 void ExpressionVisitor::visitVarExpressionNewObject(VarExpressionNewObjectAst *node)
 {
     DefaultVisitor::visitVarExpressionNewObject(node);
-    if (node->className->staticIdentifier != -1) {
+    if (node->classNameReference->className && node->classNameReference->className->staticIdentifier != -1) {
         static const QualifiedIdentifier id(QStringLiteral("static"));
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-        usingDeclaration(node->className, dec);
+        usingDeclaration(node->classNameReference->className, dec);
         m_result.setDeclaration(dec);
-    } else if (node->className->identifier) {
-        const QualifiedIdentifier id = identifierForNamespace(node->className->identifier, m_editor);
+    } else if (node->classNameReference->className && node->classNameReference->className->identifier) {
+        const QualifiedIdentifier id = identifierForNamespace(node->classNameReference->className->identifier, m_editor);
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-        usingDeclaration(node->className->identifier->namespaceNameSequence->back()->element, dec);
-        buildNamespaceUses(node->className->identifier, id);
+        usingDeclaration(node->classNameReference->className->identifier->namespaceNameSequence->back()->element, dec);
+        buildNamespaceUses(node->classNameReference->className->identifier, id);
         m_result.setDeclaration(dec);
     }
 }
@@ -216,12 +232,13 @@ void ExpressionVisitor::visitVarExpressionArray(VarExpressionArrayAst *node)
 void ExpressionVisitor::visitClosure(ClosureAst* node)
 {
     auto* closureType = new FunctionType;
-    m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeVoid)));
+    closureType->setReturnType(AbstractType::Ptr(new IntegralType(IntegralType::TypeVoid)));
+    openClosureReturnType(AbstractType::Ptr(closureType));
     if (node->functionBody) {
         visitInnerStatementList(node->functionBody);
     }
-    if (node->returnType && node->returnType->objectType) {
-        NamespacedIdentifierAst* objectType = node->returnType->objectType;
+    if (node->returnType && node->returnType->typehint && isClassTypehint(node->returnType->typehint, m_editor)) {
+        NamespacedIdentifierAst* objectType = node->returnType->typehint->genericType;
         QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
 
@@ -229,13 +246,12 @@ void ExpressionVisitor::visitClosure(ClosureAst* node)
         buildNamespaceUses(objectType, id);
     }
 
-    //First try return typehint or phpdoc return typehint
+    //Override found type with return typehint or phpdoc return typehint
     AbstractType::Ptr type = returnType(node->returnType, {}, m_editor, m_currentContext);
-    if (!type) {
-        //If failed, use the inferred type from return statements
-        type = m_result.type();
+
+    if (type) {
+        closureType->setReturnType(type);
     }
-    closureType->setReturnType(type);
 
     if (node->parameters->parametersSequence) {
         const KDevPG::ListNode< ParameterAst* >* it = node->parameters->parametersSequence->front();
@@ -243,8 +259,9 @@ void ExpressionVisitor::visitClosure(ClosureAst* node)
             AbstractType::Ptr type = parameterType(it->element, {}, m_editor, m_currentContext);
             closureType->addArgument(type);
 
-            if (it->element->parameterType && it->element->parameterType->objectType) {
-                NamespacedIdentifierAst* objectType = it->element->parameterType->objectType;
+            if (it->element->parameterType && it->element->parameterType->typehint
+                    && isClassTypehint(it->element->parameterType->typehint, m_editor)) {
+                NamespacedIdentifierAst* objectType = it->element->parameterType->typehint->genericType;
                 QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
                 DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
 
@@ -283,6 +300,7 @@ void ExpressionVisitor::visitClosure(ClosureAst* node)
     }
 
     m_result.setType(AbstractType::Ptr(closureType));
+    closeClosureReturnType();
 }
 
 void ExpressionVisitor::visitFunctionCallParameterList( FunctionCallParameterListAst* node )
@@ -353,8 +371,12 @@ void ExpressionVisitor::visitFunctionCall(FunctionCallAst* node)
             m_result.setDeclaration(dec);
         } else {
             //global function call foo();
-            const QualifiedIdentifier id = identifierForNamespace(node->stringFunctionNameOrClass, m_editor);
+            QualifiedIdentifier id = identifierForNamespace(node->stringFunctionNameOrClass, m_editor);
             DeclarationPointer dec = findDeclarationImport(FunctionDeclarationType, id);
+            if (!dec) {
+                id.setExplicitlyGlobal(true);
+                dec = findDeclarationImport(FunctionDeclarationType, id);
+            }
             ifDebug(qCDebug(DUCHAIN) << "function call of" << (dec ? dec->toString() : QString("function not found"));)
             m_result.setDeclaration(dec);
             usingDeclaration(node->stringFunctionNameOrClass->namespaceNameSequence->back()->element, dec);
@@ -439,6 +461,10 @@ void ExpressionVisitor::visitConstantOrClassConst(ConstantOrClassConstAst *node)
             //constant (created with declare('foo', 'bar')) or const Foo = 1;
             QualifiedIdentifier id = identifierForNamespace(node->constant, m_editor, true);
             DeclarationPointer declaration = findDeclarationImport(ConstantDeclarationType, id);
+            if (!declaration) {
+                id.setExplicitlyGlobal(true);
+                declaration = findDeclarationImport(ConstantDeclarationType, id);
+            }
             if (!declaration) {
                 ///TODO: is this really wanted?
                 //it could also be a global function call, without ()
@@ -622,8 +648,16 @@ void ExpressionVisitor::visitVariableProperty(VariablePropertyAst *node)
                             }
                         } else {
                             if ( !dec->isFunctionDeclaration() ) {
-                                decs << dec;
-                                ifDebug(qCDebug(DUCHAIN) << "found:" << dec->toString();)
+                                ClassMemberDeclaration *classDec = dynamic_cast<ClassMemberDeclaration*>(dec);
+                                if (classDec && classDec->accessPolicy() == Declaration::Private) {
+                                    if (declaration == dec->context()->owner()) {
+                                        decs << dec;
+                                        ifDebug(qCDebug(DUCHAIN) << "found private:" << dec->toString();)
+                                    }
+                                } else {
+                                    decs << dec;
+                                    ifDebug(qCDebug(DUCHAIN) << "found:" << dec->toString();)
+                                }
                             }
                         }
                     }
@@ -663,28 +697,152 @@ void ExpressionVisitor::visitStaticMember(StaticMemberAst* node)
 {
     //don't call DefaultVisitor::visitStaticMember(node);
     //because we would end up in visitCompoundVariableWithSimpleIndirectReference
-    if (node->variable->variable->variable) {
-        DUContext* context = findClassContext(node->className);
-        if (context) {
-            DUChainReadLocker lock(DUChain::lock());
-            m_result.setDeclarations(context->findDeclarations(identifierForNode(node->variable->variable->variable)));
-            lock.unlock();
-            if (!m_result.allDeclarations().isEmpty()) {
-                usingDeclaration(node->variable->variable->variable, m_result.allDeclarations().last());
+    if (node->staticProperty && node->staticProperty->staticProperty) {
+        if (node->staticProperty->staticProperty->variable) {
+            DUContext* context = findClassContext(node->className);
+            if (context) {
+                useDeclaration(node->staticProperty->staticProperty->variable, context);
             } else {
-                usingDeclaration(node->variable->variable->variable, DeclarationPointer());
+                usingDeclaration(node->className, DeclarationPointer());
+                m_result.setType(AbstractType::Ptr());
             }
-        } else {
-            usingDeclaration(node->className, DeclarationPointer());
+        } else if (node->staticProperty->staticProperty->expr) {
+            const QualifiedIdentifier id = identifierForNamespace(node->className, m_editor);
+            DeclarationPointer declaration = findDeclarationImport(ClassDeclarationType, id);
+            usingDeclaration(node->className->namespaceNameSequence->back()->element, declaration);
+            buildNamespaceUses(node->className, id);
+
+            visitExpr(node->staticProperty->staticProperty->expr);
+
             m_result.setType(AbstractType::Ptr());
         }
-        if (node->variable->offsetItemsSequence) {
-            const KDevPG::ListNode< DimListItemAst* >* it = node->variable->offsetItemsSequence->front();
+    }
+
+    if (node->staticProperty && node->staticProperty->offsetItemsSequence) {
+        const KDevPG::ListNode< DimListItemAst* >* it = node->staticProperty->offsetItemsSequence->front();
+        do {
+            visitDimListItem(it->element);
+        } while(it->hasNext() && (it = it->next));
+    }
+}
+
+void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
+{
+    if (node->staticProperty) {
+        DUContext* context = findClassContext(node->className->identifier);
+
+        if (context && node->staticProperty && node->staticProperty->staticProperty) {
+            if (node->staticProperty->staticProperty->variable) {
+                // static properties (object::$property)
+                useDeclaration(node->staticProperty->staticProperty->variable, context);
+            } else if (node->staticProperty->staticProperty->expr) {
+                // variable static properties (object::${$property})
+                visitExpr(node->staticProperty->staticProperty->expr);
+                usingDeclaration(node->className, DeclarationPointer());
+            }
+        }
+
+        if (node->staticProperty && node->staticProperty->offsetItemsSequence) {
+            const KDevPG::ListNode< DimListItemAst* >* dim_it = node->staticProperty->offsetItemsSequence->front();
             do {
-                visitDimListItem(it->element);
+                visitDimListItem(dim_it->element);
+            } while(dim_it->hasNext() && (dim_it = dim_it->next));
+        }
+    }
+
+    if (node->baseVariable) {
+        DefaultVisitor::visitVariableWithoutObjects(node->baseVariable);
+    }
+
+    if (node->propertiesSequence) {
+        if (!m_result.allDeclarations().isEmpty()) {
+            DUContext* context = nullptr;
+            StructureType::Ptr type;
+
+            Declaration *declaration = nullptr;
+            const KDevPG::ListNode< ClassPropertyAst* >* it = node->propertiesSequence->front();
+
+            do {
+                // first check for property names held in variables ($object->$property)
+                if (it->element->property && it->element->property->variableWithoutObjects
+                        && it->element->property->variableWithoutObjects->variable->variable) {
+                    VariableIdentifierAst *varnode = it->element->property->variableWithoutObjects->variable->variable;
+                    useDeclaration(varnode, m_currentContext);
+                } else if (it->element->property && it->element->property->variableWithoutObjects
+                        && it->element->property->variableWithoutObjects->variable->expr) {
+                        // variable dynamic properties ($object->${$property})
+                        visitExpr(it->element->property->variableWithoutObjects->variable->expr);
+                } else if (!m_result.allDeclarations().isEmpty()) {
+                    // Handle dynamic static properties first, as they don't need a class context
+                    if (it->element->staticProperty && it->element->staticProperty->staticProperty
+                      && it->element->staticProperty->staticProperty->expr) {
+                        // variable static properties ($object::${$property})
+                        visitExpr(it->element->staticProperty->staticProperty->expr);
+                        usingDeclaration(it->element->staticProperty, DeclarationPointer());
+                    }
+
+                    type = m_result.allDeclarations().last()->type<StructureType>();
+
+                    if (!type) {
+                        context = nullptr;
+                        visitClassNameReferenceDimListItems(it->element);
+                        continue;
+                    }
+
+                    DUChainReadLocker lock(DUChain::lock());
+                    declaration = type->declaration(m_currentContext->topContext());
+                    lock.unlock();
+
+                    if (!declaration) {
+                        context = nullptr;
+                        visitClassNameReferenceDimListItems(it->element);
+                        continue;
+                    }
+
+                    context = declaration->internalContext();
+
+                    if (!context || context->type() != DUContext::Class) {
+                        context = nullptr;
+                        visitClassNameReferenceDimListItems(it->element);
+                        continue;
+                    }
+
+                    if (it->element->staticProperty && it->element->staticProperty->staticProperty
+                      && it->element->staticProperty->staticProperty->variable) {
+                        // static properties ($object::$property)
+                        VariableIdentifierAst *varnode = it->element->staticProperty->staticProperty->variable;
+                        useDeclaration(varnode, context);
+                    } else if (it->element->property && it->element->property->objectDimList
+                            && it->element->property->objectDimList->variableName->name) {
+                        // normal properties ($object->property)
+                        IdentifierAst *varidnode = it->element->property->objectDimList->variableName->name;
+                        useDeclaration(varidnode, context);
+                    } else {
+                        context = nullptr;
+                    }
+
+                    visitClassNameReferenceDimListItems(it->element);
+                }
             } while(it->hasNext() && (it = it->next));
         }
     }
+
+}
+
+void ExpressionVisitor::visitClassNameReferenceDimListItems(ClassPropertyAst* node)
+{
+    // handle array indices after normal/static properties ($object->property[$index] // $object::$property[$index])
+    if (node->property && node->property->objectDimList && node->property->objectDimList->offsetItemsSequence) {
+        const KDevPG::ListNode< DimListItemAst* >* dim_it = node->property->objectDimList->offsetItemsSequence->front();
+        do {
+            visitDimListItem(dim_it->element);
+        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+    } else if (node->staticProperty && node->staticProperty->offsetItemsSequence) {
+        const KDevPG::ListNode< DimListItemAst* >* dim_it = node->staticProperty->offsetItemsSequence->front();
+        do {
+            visitDimListItem(dim_it->element);
+        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+    }
 }
 
 void ExpressionVisitor::visitUnaryExpression(UnaryExpressionAst* node)
@@ -738,11 +896,11 @@ void ExpressionVisitor::visitAdditiveExpressionRest(AdditiveExpressionRestAst* n
 void ExpressionVisitor::visitRelationalExpression(RelationalExpressionAst *node)
 {
     DefaultVisitor::visitRelationalExpression(node);
-    if (node->instanceofType && node->instanceofType->identifier) {
-        const QualifiedIdentifier id = identifierForNamespace(node->instanceofType->identifier, m_editor);
+    if (node->instanceofType && node->instanceofType->className && node->instanceofType->className->identifier) {
+        const QualifiedIdentifier id = identifierForNamespace(node->instanceofType->className->identifier, m_editor);
         DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-        usingDeclaration(node->instanceofType->identifier->namespaceNameSequence->back()->element, dec);
-        buildNamespaceUses(node->instanceofType->identifier, id);
+        usingDeclaration(node->instanceofType->className->identifier->namespaceNameSequence->back()->element, dec);
+        buildNamespaceUses(node->instanceofType->className->identifier, id);
 
         m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeBoolean)));
     }
@@ -766,6 +924,19 @@ void ExpressionVisitor::visitEqualityExpressionRest(EqualityExpressionRestAst *n
     }
 }
 
+void ExpressionVisitor::visitStatement(StatementAst *node)
+{
+    DefaultVisitor::visitStatement(node);
+
+    if (node->returnExpr) {
+        FunctionType::Ptr closureType = currentClosureReturnType().cast<FunctionType>();
+
+        if (closureType) {
+            closureType->setReturnType(m_result.type());
+        }
+    }
+}
+
 QString ExpressionVisitor::stringForNode(AstNode* id)
 {
     if (!id)
@@ -822,6 +993,30 @@ void ExpressionVisitor::buildNamespaceUses(NamespacedIdentifierAst* namespaces,
     }
 }
 
+void ExpressionVisitor::useDeclaration(VariableIdentifierAst* node, DUContext* context)
+{
+    DUChainReadLocker lock(DUChain::lock());
+    m_result.setDeclarations(context->findDeclarations(identifierForNode(node)));
+    lock.unlock();
+    if (!m_result.allDeclarations().isEmpty()) {
+        usingDeclaration(node, m_result.allDeclarations().last());
+    } else {
+        usingDeclaration(node, DeclarationPointer());
+    }
+}
+
+void ExpressionVisitor::useDeclaration(IdentifierAst* node, DUContext* context)
+{
+    DUChainReadLocker lock(DUChain::lock());
+    m_result.setDeclarations(context->findDeclarations(identifierForNode(node)));
+    lock.unlock();
+    if (!m_result.allDeclarations().isEmpty()) {
+        usingDeclaration(node, m_result.allDeclarations().last());
+    } else {
+        usingDeclaration(node, DeclarationPointer());
+    }
+}
+
 DeclarationPointer ExpressionVisitor::findDeclarationImport(DeclarationType declarationType, IdentifierAst* node)
 {
     // methods and class names are case insensitive
